@@ -34,6 +34,7 @@ G_DEFINE_INTERFACE (DzlDockItem, dzl_dock_item, GTK_TYPE_WIDGET)
 
 enum {
   MANAGER_SET,
+  PRESENTED,
   N_SIGNALS
 };
 
@@ -161,6 +162,20 @@ dzl_dock_item_default_init (DzlDockItemInterface *iface)
                   G_STRUCT_OFFSET (DzlDockItemInterface, manager_set),
                   NULL, NULL, NULL,
                   G_TYPE_NONE, 1, DZL_TYPE_DOCK_MANAGER);
+  g_signal_set_va_marshaller (signals [MANAGER_SET],
+                              G_TYPE_FROM_INTERFACE (iface),
+                              g_cclosure_marshal_VOID__OBJECTv);
+
+  signals [PRESENTED] =
+    g_signal_new ("presented",
+                  G_TYPE_FROM_INTERFACE (iface),
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (DzlDockItemInterface, presented),
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
+  g_signal_set_va_marshaller (signals [PRESENTED],
+                              G_TYPE_FROM_INTERFACE (iface),
+                              g_cclosure_marshal_VOID__VOIDv);
 }
 
 /**
@@ -357,8 +372,18 @@ dzl_dock_item_present (DzlDockItem *self)
     {
       if (DZL_IS_DOCK_ITEM (parent))
         {
+          DzlDockManager *manager;
+
           dzl_dock_item_present_child (DZL_DOCK_ITEM (parent), self);
           dzl_dock_item_present (DZL_DOCK_ITEM (parent));
+
+          /* gtk_widget_grab_focus() results in a transient grab,
+           * we want a real grab (that doesn't release the parent)
+           * when we are focused during this code path.
+           */
+          if ((manager = dzl_dock_item_get_manager (self)))
+            dzl_dock_manager_release_transient_grab (manager);
+
           return;
         }
     }
@@ -661,3 +686,26 @@ dzl_dock_item_get_can_minimize (DzlDockItem *self)
 
   return FALSE;
 }
+
+/**
+ * dzl_dock_item_emit_presented:
+ * @self: a #DzlDockItem
+ *
+ * Emits the #DzlDockItem::presented signal.
+ *
+ * Containers should emit this when their descendant has been presented as the
+ * current visible child. This allows dock items to do lazy initialization of
+ * content once the widgetry is visible to the user.
+ *
+ * Currently, this is best effort, as there are a number of situations that
+ * make covering all cases problematic.
+ *
+ * Since: 3.30
+ */
+void
+dzl_dock_item_emit_presented (DzlDockItem *self)
+{
+  g_return_if_fail (DZL_IS_DOCK_ITEM (self));
+
+  g_signal_emit (self, signals [PRESENTED], 0);
+}
