@@ -29,6 +29,7 @@
 #include "suggestions/dzl-suggestion-entry-buffer.h"
 #include "suggestions/dzl-suggestion-popover.h"
 #include "suggestions/dzl-suggestion-private.h"
+#include "util/dzl-gtk.h"
 #include "util/dzl-util-private.h"
 
 typedef struct
@@ -59,6 +60,7 @@ enum {
 };
 
 enum {
+  ACTION,
   ACTIVATE_SUGGESTION,
   HIDE_SUGGESTIONS,
   MOVE_SUGGESTION,
@@ -139,15 +141,35 @@ dzl_suggestion_entry_hide_suggestions (DzlSuggestionEntry *self)
   DZL_EXIT;
 }
 
+static gboolean
+dzl_suggestion_entry_focus_in_event (GtkWidget     *widget,
+                                     GdkEventFocus *event)
+{
+  DzlSuggestionEntry *self = (DzlSuggestionEntry *)widget;
+  DzlSuggestionEntryPrivate *priv = dzl_suggestion_entry_get_instance_private (self);
+
+  g_assert (DZL_IS_SUGGESTION_ENTRY (self));
+  g_assert (event != NULL);
+
+  if (priv->popover)
+    _dzl_suggestion_popover_set_focused (priv->popover, TRUE);
+
+  return GTK_WIDGET_CLASS (dzl_suggestion_entry_parent_class)->focus_in_event (widget, event);
+}
+
 static gboolean
 dzl_suggestion_entry_focus_out_event (GtkWidget     *widget,
                                       GdkEventFocus *event)
 {
   DzlSuggestionEntry *self = (DzlSuggestionEntry *)widget;
+  DzlSuggestionEntryPrivate *priv = dzl_suggestion_entry_get_instance_private (self);
 
   g_assert (DZL_IS_SUGGESTION_ENTRY (self));
   g_assert (event != NULL);
 
+  if (priv->popover)
+    _dzl_suggestion_popover_set_focused (priv->popover, FALSE);
+
   g_signal_emit (self, signals [HIDE_SUGGESTIONS], 0);
 
   return GTK_WIDGET_CLASS (dzl_suggestion_entry_parent_class)->focus_out_event (widget, event);
@@ -177,11 +199,18 @@ dzl_suggestion_entry_key_press_event (GtkWidget   *widget,
 {
   DzlSuggestionEntry *self = (DzlSuggestionEntry *)widget;
   DzlSuggestionEntryPrivate *priv = dzl_suggestion_entry_get_instance_private (self);
+  GdkDevice *device;
   gboolean ret;
 
   g_assert (DZL_IS_SUGGESTION_ENTRY (self));
   g_assert (priv->in_key_press >= 0);
 
+  if ((device = gdk_event_get_device ((GdkEvent *)key)) &&
+      gdk_device_get_source (device) == GDK_SOURCE_KEYBOARD)
+    device = gdk_device_get_associated_device (device);
+
+  _dzl_suggestion_popover_set_device (priv->popover, device);
+
   /*
    * If Tab was pressed, and there is uncommitted suggested text,
    * commit it and stop propagation of the key press.
@@ -332,10 +361,11 @@ dzl_suggestion_entry_move_suggestion (DzlSuggestionEntry *self,
 
   g_assert (DZL_IS_SUGGESTION_ENTRY (self));
 
+  if (priv->popover == NULL || !gtk_widget_get_mapped (GTK_WIDGET (priv->popover)))
+    return;
+
   priv->in_move_by++;
-
   dzl_suggestion_popover_move_by (priv->popover, amount);
-
   priv->in_move_by--;
 }
 
@@ -479,6 +509,7 @@ dzl_suggestion_entry_class_init (DzlSuggestionEntryClass *klass)
   object_class->set_property = dzl_suggestion_entry_set_property;
 
   widget_class->destroy = dzl_suggestion_entry_destroy;
+  widget_class->focus_in_event = dzl_suggestion_entry_focus_in_event;
   widget_class->focus_out_event = dzl_suggestion_entry_focus_out_event;
   widget_class->hierarchy_changed = dzl_suggestion_entry_hierarchy_changed;
   widget_class->key_press_event = dzl_suggestion_entry_key_press_event;
@@ -534,6 +565,18 @@ dzl_suggestion_entry_class_init (DzlSuggestionEntryClass *klass)
 
   g_object_class_install_properties (object_class, N_PROPS, properties);
 
+  signals [ACTION] =
+    g_signal_new_class_handler ("action",
+                                G_TYPE_FROM_CLASS (klass),
+                                G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+                                G_CALLBACK (dzl_gtk_widget_action_with_string),
+                                NULL, NULL, NULL,
+                                G_TYPE_NONE,
+                                3,
+                                G_TYPE_STRING,
+                                G_TYPE_STRING,
+                                G_TYPE_STRING);
+
   signals [HIDE_SUGGESTIONS] =
     g_signal_new ("hide-suggestions",
                   G_TYPE_FROM_CLASS (klass),
@@ -638,6 +681,7 @@ dzl_suggestion_entry_init (DzlSuggestionEntry *self)
                                 "modal", FALSE,
                                 "relative-to", self,
                                 "type", GTK_WINDOW_POPUP,
+                                "type-hint", GDK_WINDOW_TYPE_HINT_COMBO,
                                 NULL);
   g_signal_connect_object (priv->popover,
                            "notify::selected",
@@ -754,7 +798,8 @@ dzl_suggestion_entry_set_model (DzlSuggestionEntry *self,
       g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_MODEL]);
       dzl_suggestion_entry_update_attrs (self);
 
-      g_signal_emit (self, signals [SHOW_SUGGESTIONS], 0);
+      if (gtk_widget_has_focus (GTK_WIDGET (self)))
+        g_signal_emit (self, signals [SHOW_SUGGESTIONS], 0);
     }
 
   DZL_EXIT;
@@ -984,3 +1029,24 @@ _dzl_suggestion_entry_reposition (DzlSuggestionEntry   *self,
   gtk_widget_set_size_request (GTK_WIDGET (popover), alloc.width, -1);
   gtk_window_move (GTK_WINDOW (popover), alloc.x, alloc.y);
 }
+
+/**
+ * dzl_suggestion_entry_get_popover:
+ * @self: a #DzlSuggestionEntry
+ *
+ * Gets the result display widget for the entry. This is currently
+ * always a #DzlSuggestionPopover.
+ *
+ * Returns: (transfer none): a #GtkWidget
+ *
+ * Since: 3.32
+ */
+GtkWidget *
+dzl_suggestion_entry_get_popover (DzlSuggestionEntry *self)
+{
+  DzlSuggestionEntryPrivate *priv = dzl_suggestion_entry_get_instance_private (self);
+
+  g_return_val_if_fail (DZL_IS_SUGGESTION_ENTRY (self), NULL);
+
+  return GTK_WIDGET (priv->popover);
+}
